// 東芝RDシリーズ DVD／HDDレコーダの“ネットdeダビング”機能に特化したFTPサーバの簡易的な実装

//  Created by raktajino on Sun Nov. 21 2004.
//  Copyright (c) 2004 raktajino. All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package com.mac.raktajino.RDService;

import java.io.*;
import java.net.*;
import java.util.StringTokenizer;
import java.text.NumberFormat;
import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;

public class FTPSession extends Thread {

	protected 				Socket	fControlSocket;
	protected				int		fFTPDataPortFrom;
	protected				int		fFTPDataPortTo;
	protected				int 	fFTPBufferSize;
	protected 				String	fDstDirectory;

	// OSから情報を取得できなかった場合に使用するディスク情報のデフォルト値。
	protected static final	long	kDefaultFreeSpace = 600 * 1024L * 1024L * 1024L;
	protected static final	long	kDefaultUsedSpace =   0 * 1024L * 1024L * 1024L;

	// --- コンストラクタ

	FTPSession(Socket controlSocket, int dataPortFrom, int dataPortTo, int bufferSize, String dstDirectory)
	{
		fControlSocket   = controlSocket;
		fFTPDataPortFrom = dataPortFrom;
		fFTPDataPortTo   = dataPortTo;
		fFTPBufferSize   = bufferSize;
		fDstDirectory    = dstDirectory;
	}
	
	// --- スレッドの本体を実行する。
	
	public void run()
	{
		try {
			this.doFTPSession();
		}
        catch(Exception e) {
            System.err.println("FTP session : " + e.getMessage());
        }
	}

	// --- 1回分のFTPセッションを処理する。
	
	public void doFTPSession()
		throws IOException, ParserConfigurationException, SAXException
	{
		PrintWriter    controlWriter = null;
		BufferedReader controlReader = null;
		ServerSocket   ftpDataSocket = null;
		try {
			// 指定されたポート番号の範囲で、データ転送用のソケットの作成を試みる。
			for(int dataPort = fFTPDataPortFrom; dataPort <= fFTPDataPortTo; dataPort++) {
				try {
					ftpDataSocket = new ServerSocket(dataPort);
					break;
				}
				catch(IOException e) {
					if(dataPort == fFTPDataPortTo) throw e;
				}
			}

			controlWriter = new PrintWriter(fControlSocket.getOutputStream());
			controlReader = new BufferedReader(new InputStreamReader(fControlSocket.getInputStream()));
			this.writeStatus(controlWriter, "220 Connected. <00:00:00:000>.");
	
			long titleSize = -1;
			while(true) {
				String commandLine = controlReader.readLine();
				if(commandLine == null) break;
	
	    		System.out.println(commandLine);
				
				StringTokenizer tokenizer = new StringTokenizer(commandLine);
				String command = tokenizer.nextToken();
	
				if(command.equals("USER")) {
					// ユーザ名を指定し、ログインする。
					// 簡易実装なので、任意のユーザ名でログイン可能とする。
					this.writeStatus(controlWriter, "331 Anonymous login OK, send e-mail as password.");
				}
				else if(command.equals("PASS")) {
					// パスワードを指定する。
					// 簡易実装なので、任意のパスワードでログイン可能とする。
					this.writeStatus(controlWriter, "230 Anonymous user logged in, proceed.");
				}
				else if(command.equals("TYPE")) {
					// 転送データのタイプを指定する。
					// 簡易実装なので、無条件に“I”を設定する。
					this.writeStatus(controlWriter, "200 Type set to I.");
				}
				else if(command.equals("PASV")) {
					// パッシブ・モードを指定する。
					String localHostAddress = fControlSocket.getLocalAddress().getHostAddress().replace('.', ',');
					int    dataPortNumber   = ftpDataSocket.getLocalPort();
					String dataPortString   = String.valueOf(dataPortNumber / 256) + "," + String.valueOf(dataPortNumber % 256);
					this.writeStatus(controlWriter, "227 Entering Passive Mode (" + localHostAddress + "," + dataPortString + ").");
				}
				else if(command.equals("STOR")) {
					// RDからのファイルを受信する。
					if(!tokenizer.hasMoreElements()) {
						this.writeStatus(controlWriter, "501 Syntax error in parameters or arguments.");
					}
					else {
						String      fileName      = this.copySafeFileName(tokenizer.nextToken());
						InetAddress remoteAddress = fControlSocket.getInetAddress();
						long        size          = this.receiveFile(fileName, ftpDataSocket, remoteAddress, controlWriter);
						if(size >= 0) {
							titleSize = size;
						}
					}
				}
				else if(command.equals("RETR")) {
					// RDへファイルを送信する。
					if(!tokenizer.hasMoreElements()) {
						this.writeStatus(controlWriter, "501 Syntax error in parameters or arguments.");
					}
					else {
						String      fileName      = tokenizer.nextToken();
						InetAddress remoteAddress = fControlSocket.getInetAddress();
						this.sendFile(fileName, ftpDataSocket, remoteAddress, controlWriter, titleSize);
					}
				}
				else if(command.equals("QUIT")) {
					// FTPセッションを終了する。
					this.writeStatus(controlWriter, "221 Thank you for using the FTP service.");
					break;
				}
				else {
					// その他のコマンドは未実装とする。
					this.writeStatus(controlWriter, "502 Command not implemented.");
				}
			}
		}
		finally {
			if(controlWriter  != null) controlWriter.close();
			if(controlReader  != null) controlReader.close();
			if(ftpDataSocket  != null) ftpDataSocket.close();
			if(fControlSocket != null) fControlSocket.close();
		}
	}

	// --- RDからのファイルを受信する。
	
	protected long receiveFile(String fileName, ServerSocket ftpDataSocket, InetAddress remoteAddress, PrintWriter controlWriter)
		throws IOException, ParserConfigurationException, SAXException
	{
		long titleSize = -1;	

		// ファイル名とその拡張子を決定する。
		String dstFileExtension = null;
		if(fileName.equals("$netdubbing$dev0.dat")) {
			dstFileExtension = ".dat";
		}
		else if(fileName.equals("$netdubbing$dubbinginfo.xml")) {
			dstFileExtension = ".xml";
		}
		else {
			// セキュリティ上の観点から、RDから転送されるはずのファイル以外はすべてエラーとする。
			this.writeStatus(controlWriter, "550 Requested action not taken.");
			return titleSize;
		}
	
		// データ転送用のソケットへのコネクションを待ち受ける。
		Socket dataSocket = ftpDataSocket.accept();
		if(dataSocket == null) {
			return titleSize;
		}
		if(!dataSocket.getInetAddress().equals(remoteAddress)) {
			dataSocket.close();
			return titleSize;
		}
		InputStream dataStream = dataSocket.getInputStream();
		this.writeStatus(controlWriter, "150 File status okay; about to open data connection.");

		// ファイルを受信する。
		File             outputFile   = null;
		FileOutputStream outputStream = null;
		try {
		
			// --- $netdubbing$dev0.datファイルの書き込み先ファイルを用意する。
		
			if(dstFileExtension.equals(".dat")) {
				// 番組のタイトルを読み、ファイル名を決定する。
				byte[] header      = new byte [0x1800];
				int    headerIndex = 0;
				while(headerIndex < header.length) {
					int headerBytes = dataStream.read(header, headerIndex, header.length - headerIndex);
					if(headerBytes < 0) {
						this.writeStatus(controlWriter, "550 Requested action not taken.");
						return titleSize;
					}
					headerIndex += headerBytes;
				}
				String programTitle = this.copySafeFileName(this.readProgramTitle(header));
				if(programTitle == null) {
					this.writeStatus(controlWriter, "550 Requested action not taken.");
					return titleSize;
				}
				
				// 書き込み先のファイルを作成する。
				outputFile   = this.createUniqueFile(fDstDirectory, programTitle, ".mpg");
				outputStream = new FileOutputStream(outputFile);

				// 書き込み先のファイルのパーミッションを（可能ならば）設定する。
				try {
					String[] cmdArray = { "/bin/chmod", "a+w", outputFile.getAbsolutePath() };
					Process  process  = Runtime.getRuntime().exec(cmdArray);
		    		process.waitFor();
				}
				catch(IOException e) {
				}
				catch(InterruptedException e) {
				}

				// 書き込み先ファイルの名前を表示する。
				System.out.println(outputFile.getName());

				// 先読みしたヘッダ部分をファイルへ書き込む。
				outputStream.write(header);
			}

			// --- $netdubbing$dubbinginfo.xmlファイルの書き込み先ファイルを用意する。
			
			else if(dstFileExtension.equals(".xml")) {
				try {
					outputFile = File.createTempFile(Main.kProgramName, dstFileExtension);
					outputFile.deleteOnExit();
					outputStream = new FileOutputStream(outputFile);
				}
				catch(NoSuchMethodError e) {
					// JDK 1.2より古い環境では、createTempFileメソッドを使用できない。
					// createTempFileメソッドを使わずにファイルを作る。
					outputFile   = this.createUniqueFile(fDstDirectory, "dubbinginfo", ".xml");
					outputStream = new FileOutputStream(outputFile);
				}
			}
			
			// --- その他のファイルはエラーとする。（通常、起こりえない。）
			
			else {
				this.writeStatus(controlWriter, "550 Requested action not taken.");
				return titleSize;
			}
			
			// ファイルを受信する。
			try {
				byte[] buffer = new byte[fFTPBufferSize];
				while(true) {
					int readBytes = dataStream.read(buffer);
					if(readBytes < 0) break;
					if(outputStream != null) {
						outputStream.write(buffer, 0, readBytes);
					}
				}
			}
			finally {
				if(outputStream != null) {
					outputStream.close();
				}
			}

			// ダビングするファイルのバイト数を取得する。
			if(dstFileExtension.equals(".xml") && outputFile != null) {
				titleSize = this.readTitleSize(outputFile);
			}
		}
		finally {
			dataSocket.close();
			if(dstFileExtension.equals(".xml") && outputFile != null) {
				outputFile.delete();
			}
		}
		this.writeStatus(controlWriter, "226 Closing data connection.");
		return titleSize;
	}

	// --- RDへファイルを送信する。

	protected void sendFile(String fileName, ServerSocket ftpDataSocket, InetAddress remoteAddress, PrintWriter controlWriter, long titleSize)
		throws IOException
	{
		// ファイル名をチェックする。
		// セキュリティ上の観点から、RDから要求されるはずのファイル以外はすべてエラーとする。
		if(!fileName.equals("$xml$status.xml")) {
			this.writeStatus(controlWriter, "550 " + fileName + ": No such file or directory.");
			return;
		}

		// データ転送用のソケットへのコネクションを待ち受ける。
		Socket dataSocket = ftpDataSocket.accept();
		if(dataSocket == null) return;
		if(!dataSocket.getInetAddress().equals(remoteAddress)) {
			dataSocket.close();
			return;
		}

		// ファイルの送信を行う。
		try {
			this.writeStatus(controlWriter, "150 File status okay; about to open data connection.");

			// ディスクの空き容量と使用容量を求める。
			DiskInformation diskInformation = new DiskInformation(new File(fDstDirectory));
			long freeSpace = diskInformation.fFreeSpace;
			long usedSpace = diskInformation.fUsedSpace;
			if(freeSpace < 0) freeSpace = kDefaultFreeSpace;
			if(usedSpace < 0) usedSpace = kDefaultUsedSpace;
			
			NumberFormat numberFormat = NumberFormat.getInstance();
			System.out.println(numberFormat.format(usedSpace) + " bytes used, " + numberFormat.format(freeSpace) + " bytes free.");
			if(titleSize >= 0) {
				System.out.println(numberFormat.format(titleSize) + " bytes will be received.");
			}
			
			// ステータスをRDヘ送信する。
			PrintWriter dataWriter = new PrintWriter(dataSocket.getOutputStream());
			dataWriter.println("<?xml version=\"1.0\" encoding=\"shift_jis\"?>");
			dataWriter.println("<status>");
			dataWriter.println("<dubbing>ready</dubbing>");
			if(titleSize < 0 || titleSize <= freeSpace) {
				dataWriter.println("<accept_dubbing>ready</accept_dubbing>");
			}
			else {
				dataWriter.println("<accept_dubbing>busy</accept_dubbing>");
				System.out.println("The dubbing will be stopped, because there is not enough space on the disk.");
			}
			
			// RDへは2048バイト単位で容量を送信する。
			dataWriter.println("<device0_remain>" + freeSpace / 2048L + "</device0_remain>");
			dataWriter.println("<device0_recorded>" + usedSpace / 2048L + "</device0_recorded>");
			dataWriter.println("<device0_title_remain>362</device0_title_remain>");
			dataWriter.println("</status>");
			dataWriter.close();
		}
		finally {
			dataSocket.close();
		}
		this.writeStatus(controlWriter, "226 Closing data connection.");
	}

	// --- FTPのステータスを送信する。
	
	protected void writeStatus(PrintWriter writer, String status)
	{
		writer.println(status);
		writer.flush();
		System.out.println(status);
	}
	
	// --- MPEG2データの先頭部分より、番組のタイトルを取得する。
	
	protected String readProgramTitle(byte[] header)
		throws IOException
	{
		String title  = null;
		if(header.length >= 0x84 + 64) {
			title = (new String(header, 0x84, 64, "Shift_JIS")).trim();
		}
		return title;
	}

	// --- ファイル名から危険な文字を取り除く。
	
	protected String copySafeFileName(String rawName)
	{	
		StringBuffer buffer = new StringBuffer(rawName.trim());	// 前後の空白や制御文字は取り除く。
		
		for(int n = 0; n < buffer.length(); n++) {
			char ch = buffer.charAt(n);
			if(ch == File.separatorChar) {
				// パスの区切り文字（UnixやMacOS Xでは、通常、スラッシュ）を、安全な文字に置き換える。
				buffer.setCharAt(n, '_');
			}
			else if(ch == '.') {
				// 2個以上の連続するドットを、安全な文字に置き換える。
				int dotCount;
				for(dotCount = 1 ; n + dotCount < buffer.length(); dotCount++) {
					if(buffer.charAt(n + dotCount) != '.') break;
				}
				if(dotCount >= 2) {
					buffer.setCharAt(n, '_');
					buffer.delete(n + 1, n + dotCount);
				}
			}
			else if(Main.kPlatformMacOS && (ch == ':')) {
				// MacOSで使用不可の文字を、安全な文字に置き換える。
				buffer.setCharAt(n, '_');
			}
			else if(Main.kPlatformWindows && (ch == '\\' || ch == '/' || ch == ':' || ch == '?' ||
											  ch == '*'  || ch == '"' || ch == '>' || ch == '<' || ch == '|')) {
				// Windowsで使用不可の文字を、安全な文字に置き換える。
				buffer.setCharAt(n, '_');
			}
		}
		return buffer.toString();
	}
	
	// --- ユニークな名前のファイルを作る。
	
	protected synchronized File createUniqueFile(String directory, String bodyName, String extensionName)
		throws IOException
	{
		File file           = new File(directory, bodyName + extensionName);
		int  sequenceNumber = 1;
		while(file.exists()) {
			// 同名のファイルが既に存在していたら、末尾に連番を付けて重複を回避する。
			file = new File(directory, bodyName + "#" + sequenceNumber++ + extensionName);
		}
		
		// 空のファイルを作成する。
		new FileOutputStream(file).close();
		return file;
	}

	// --- $netdubbing$dubbinginfo.xmlファイルより、ダビングするファイルのバイト数を取得する。
	
	protected long readTitleSize(File infoFile)
		throws IOException, ParserConfigurationException, SAXException
	{
		long titleSize = -1;
		try {
			DocumentBuilderFactory factory       = DocumentBuilderFactory.newInstance();
			DocumentBuilder        builder       = factory.newDocumentBuilder();
			Document               document      = builder.parse(infoFile);
			Element                root          = document.getDocumentElement();
			NodeList               deviceList    = root.getElementsByTagName("titlesize");
	        Element                deviceElement = (Element)deviceList.item(0);
			titleSize = Long.parseLong(deviceElement.getFirstChild().getNodeValue()) * 2048L;
		}
		catch(NoClassDefFoundError e) {
			titleSize = -1;
		}
		catch(NoSuchMethodError e) {
			titleSize = -1;
		}
		return titleSize;
	}
}
